@@ -1,27 +1,20 @@
-use core::{cell::UnsafeCell, mem::MaybeUninit, slice};
+use core::{arch::asm, cell::UnsafeCell, mem::MaybeUninit, slice};
 use memoffset::offset_of;
-use r3::{
-    kernel::{Port, PortToKernel, TaskCb},
-    prelude::*,
-    utils::{Init, ZeroInit},
-};
-use r3_portkit::sym::{sym_static, SymStaticExt};
+use r3_core::{kernel::traits, utils::Init};
+use r3_kernel::{KernelTraits, Port, PortToKernel, System, TaskCb};
+use r3_portkit::sym::sym_static;
 
 use super::cfg::{InterruptController, ThreadingOptions, Timer};
 
-/// Implemented on a system type by [`use_port!`].
+/// Implemented on a kernel trait type by [`use_port!`].
 ///
 /// # Safety
 ///
 /// Only meant to be implemented by [`use_port!`].
 pub unsafe trait PortInstance:
-    Kernel + Port<PortTaskState = TaskState> + ThreadingOptions + InterruptController + Timer
+    KernelTraits + Port<PortTaskState = TaskState> + ThreadingOptions + InterruptController + Timer
 {
-    sym_static!(static PORT_STATE: SymStatic<State> = zeroed!());
-
-    fn port_state() -> &'static State {
-        sym_static(Self::PORT_STATE).as_ref()
-    }
+    sym_static!(#[sym(p_port_state)] fn port_state() -> &State);
 }
 
 #[repr(C)]
@@ -41,7 +34,14 @@ impl State {
 }
 
 unsafe impl Sync for State {}
-unsafe impl ZeroInit for State {}
+
+impl Init for State {
+    const INIT: Self = Self {
+        dispatch_pending: UnsafeCell::new(false),
+        main_stack: UnsafeCell::new(0),
+        running_task_ptr: UnsafeCell::new(core::ptr::null_mut()),
+    };
+}
 
 #[derive(Debug)]
 #[repr(C)]
@@ -52,29 +52,32 @@ pub struct TaskState {
 unsafe impl Sync for TaskState {}
 
 impl Init for TaskState {
+    #[allow(clippy::declare_interior_mutable_const)] // it's intentional
     const INIT: Self = Self {
         sp: UnsafeCell::new(0),
     };
 }
 
 impl State {
-    pub unsafe fn port_boot<System: PortInstance>(&self) -> ! {
-        unsafe { self.enter_cpu_lock::<System>() };
+    #[inline(always)]
+    pub unsafe fn port_boot<Traits: PortInstance>(&self) -> ! {
+        unsafe { self.enter_cpu_lock::<Traits>() };
 
-        unsafe { *self.running_task_ptr.get() = System::state().running_task_ptr() as *mut () };
+        unsafe { *self.running_task_ptr.get() = Traits::state().running_task_ptr() as *mut () };
 
         // Safety: We are the port, so it's okay to call this
-        unsafe { <System as InterruptController>::init() };
+        unsafe { <Traits as InterruptController>::init() };
 
         // Safety: We are the port, so it's okay to call this
-        unsafe { <System as Timer>::init() };
+        unsafe { <Traits as Timer>::init() };
 
         // Safety: We are the port, so it's okay to call this
-        unsafe { <System as PortToKernel>::boot() };
+        unsafe { <Traits as PortToKernel>::boot() };
     }
 
-    pub unsafe fn dispatch_first_task<System: PortInstance>(&'static self) -> ! {
-        debug_assert!(self.is_cpu_lock_active::<System>());
+    #[inline(always)]
+    pub unsafe fn dispatch_first_task<Traits: PortInstance>(&'static self) -> ! {
+        debug_assert!(self.is_cpu_lock_active::<Traits>());
 
         unsafe {
             asm!("
@@ -87,12 +90,13 @@ impl State {
                 mov sp, r0
 
                 # Save the stack pointer for later use
+                # [tag:arm_main_stack_assigned_in_dft]
                 str r0, [r1]
 
                 b {push_second_level_state_and_dispatch}.dispatch
                 ",
                 push_second_level_state_and_dispatch =
-                    sym Self::push_second_level_state_and_dispatch::<System>,
+                    sym Self::push_second_level_state_and_dispatch::<Traits>,
                 in("r1") self.main_stack.get(),
                 options(noreturn),
             );
@@ -100,8 +104,8 @@ impl State {
     }
 
     #[inline(never)] // avoid symbol collision with `YieldReturn`
-    pub unsafe fn yield_cpu<System: PortInstance>(&'static self) {
-        if !self.is_task_context::<System>() {
+    pub unsafe fn yield_cpu<Traits: PortInstance>(&'static self) {
+        if !self.is_task_context::<Traits>() {
             unsafe { self.dispatch_pending.get().write_volatile(true) };
             return;
         }
@@ -134,7 +138,7 @@ impl State {
             0:        # YieldReturn
                 ",
                 push_second_level_state_and_dispatch =
-                    sym Self::push_second_level_state_and_dispatch::<System>,
+                    sym Self::push_second_level_state_and_dispatch::<Traits>,
                 out("r2") _,
                 out("r3") _,
             );
@@ -148,7 +152,7 @@ impl State {
     ///     - Push the second-level state.
     ///     - Store SP to the current task's `TaskState`.
     ///  - **`dispatch:`** (alternate entry point)
-    ///     - Call [`r3::kernel::PortToKernel::choose_running_task`].
+    ///     - Call [`r3_kernel::PortToKernel::choose_running_task`].
     ///     - Restore SP from the next scheduled task's `TaskState`.
     ///  - If there's no task to schedule, branch to [`Self::idle_task`].
     ///  - Pop the second-level state of the next scheduled task.
@@ -164,19 +168,20 @@ impl State {
     ///  - This function may overwrite any contents in the main stack.
     ///
     #[naked]
-    unsafe extern "C" fn push_second_level_state_and_dispatch<System: PortInstance>() -> ! {
-        extern "C" fn choose_and_get_next_task<System: PortInstance>(
-        ) -> Option<&'static TaskCb<System>> {
+    unsafe extern "C" fn push_second_level_state_and_dispatch<Traits: PortInstance>() -> ! {
+        extern "C" fn choose_and_get_next_task<Traits: PortInstance>(
+        ) -> Option<&'static TaskCb<Traits>> {
             // Safety: CPU Lock active
-            unsafe { System::choose_running_task() };
+            unsafe { Traits::choose_running_task() };
 
-            unsafe { *System::state().running_task_ptr() }
+            unsafe { *Traits::state().running_task_ptr() }
         }
 
         unsafe {
             asm!("
-                movw r0, :lower16:{PORT_STATE}_
-                movt r0, :upper16:{PORT_STATE}_
+                movw r0, :lower16:{p_port_state}_
+                movt r0, :upper16:{p_port_state}_
+                ldr r0, [r0]
 
                 # Skip saving the second-level state if the current context
                 # is an idle task. Also, in this case, we don't have a stack,
@@ -200,8 +205,8 @@ impl State {
 
                 # Store SP to `TaskState`.
                 #
-                #    <r0 = &PORT_STATE>
-                #    r0 = *PORT_STATE.running_task_ptr // == running_task
+                #    <r0 = &port_state>
+                #    r0 = *port_state.running_task_ptr // == running_task
                 #    r0.port_task_state.sp = sp_usr
                 #
                 ldr r0, [r0, #{OFFSET_RUNNING_TASK_PTR}]
@@ -251,11 +256,11 @@ impl State {
                 pop {{r0-r3, r12, lr}}
                 rfeia sp!
             ",
-                choose_and_get_next_task = sym choose_and_get_next_task::<System>,
+                choose_and_get_next_task = sym choose_and_get_next_task::<Traits>,
                 push_second_level_state_and_dispatch =
-                    sym Self::push_second_level_state_and_dispatch::<System>,
-                idle_task = sym Self::idle_task::<System>,
-                PORT_STATE = sym System::PORT_STATE,
+                    sym Self::push_second_level_state_and_dispatch::<Traits>,
+                idle_task = sym Self::idle_task::<Traits>,
+                p_port_state = sym Traits::p_port_state,
                 OFFSET_RUNNING_TASK_PTR = const Self::OFFSET_RUNNING_TASK_PTR,
                 OFFSET_MAIN_STACK = const Self::OFFSET_MAIN_STACK,
                 options(noreturn),
@@ -267,14 +272,15 @@ impl State {
     /// is set. Otherwise, branch to `pop_first_level_state` (thus skipping the
     /// saving/restoration of second-level states).
     #[naked]
-    unsafe extern "C" fn push_second_level_state_and_dispatch_shortcutting<System: PortInstance>(
+    unsafe extern "C" fn push_second_level_state_and_dispatch_shortcutting<Traits: PortInstance>(
     ) -> ! {
         unsafe {
             asm!("
                 # Read `port_state().dispatch_pending`. If it's clear, branch
                 # to `NotShortcutting`
-                movw r0, :lower16:{PORT_STATE}_
-                movt r0, :upper16:{PORT_STATE}_
+                movw r0, :lower16:{p_port_state}_
+                movt r0, :upper16:{p_port_state}_
+                ldr r0, [r0]
                 ldrb r1, [r0, #{OFFSET_DISPATCH_PENDING}]
                 tst r1, r1
                 bne 0f
@@ -300,9 +306,9 @@ impl State {
                 b {push_second_level_state_and_dispatch}
             ",
                 push_second_level_state_and_dispatch =
-                    sym Self::push_second_level_state_and_dispatch::<System>,
-                idle_task = sym Self::idle_task::<System>,
-                PORT_STATE = sym System::PORT_STATE,
+                    sym Self::push_second_level_state_and_dispatch::<Traits>,
+                idle_task = sym Self::idle_task::<Traits>,
+                p_port_state = sym Traits::p_port_state,
                 OFFSET_DISPATCH_PENDING = const Self::OFFSET_DISPATCH_PENDING,
                 options(noreturn),
             );
@@ -317,10 +323,10 @@ impl State {
     /// # Safety
     ///
     ///  - The processor should be in System mode (task context).
-    ///  - `*System::state().running_task_ptr()` should be `None`.
+    ///  - `*Traits::state().running_task_ptr()` should be `None`.
     ///
     #[naked]
-    unsafe extern "C" fn idle_task<System: PortInstance>() -> ! {
+    unsafe extern "C" fn idle_task<Traits: PortInstance>() -> ! {
         unsafe {
             asm!(
                 "
@@ -338,9 +344,10 @@ impl State {
         }
     }
 
-    pub unsafe fn exit_and_dispatch<System: PortInstance>(
+    #[inline(always)]
+    pub unsafe fn exit_and_dispatch<Traits: PortInstance>(
         &'static self,
-        _task: &'static TaskCb<System>,
+        _task: &'static TaskCb<Traits>,
     ) -> ! {
         unsafe {
             asm!(
@@ -349,26 +356,26 @@ impl State {
                 b {push_second_level_state_and_dispatch}.dispatch
                 ",
                 push_second_level_state_and_dispatch =
-                    sym Self::push_second_level_state_and_dispatch::<System>,
+                    sym Self::push_second_level_state_and_dispatch::<Traits>,
                 options(noreturn),
             );
         }
     }
 
     #[inline(always)]
-    pub unsafe fn enter_cpu_lock<System: PortInstance>(&self) {
+    pub unsafe fn enter_cpu_lock<Traits: PortInstance>(&self) {
         // TODO: support unmanaged interrupts
         unsafe { asm!("cpsid i") };
     }
 
     #[inline(always)]
-    pub unsafe fn leave_cpu_lock<System: PortInstance>(&'static self) {
+    pub unsafe fn leave_cpu_lock<Traits: PortInstance>(&'static self) {
         unsafe { asm!("cpsie i") };
     }
 
-    pub unsafe fn initialize_task_state<System: PortInstance>(
+    pub unsafe fn initialize_task_state<Traits: PortInstance>(
         &self,
-        task: &'static TaskCb<System>,
+        task: &'static TaskCb<Traits>,
     ) {
         let stack = task.attr.stack.as_ptr();
         let mut sp = (stack as *mut u8).wrapping_add(stack.len()) as *mut MaybeUninit<u32>;
@@ -384,7 +391,7 @@ impl State {
         };
 
         // R0: Parameter to the entry point
-        first_level[0] = MaybeUninit::new(task.attr.entry_param as u32);
+        first_level[0] = unsafe { core::mem::transmute(task.attr.entry_param) };
         // R1-R3, R12: Uninitialized
         if preload_all {
             first_level[1] = MaybeUninit::new(0x01010101);
@@ -393,7 +400,8 @@ impl State {
             first_level[4] = MaybeUninit::new(0x12121212);
         }
         // LR: The return address
-        first_level[5] = MaybeUninit::new(System::exit_task as usize as u32);
+        first_level[5] =
+            MaybeUninit::new(<System<Traits> as traits::KernelBase>::raw_exit_task as usize as u32);
         // PC: The entry point
         first_level[6] = MaybeUninit::new(task.attr.entry_point as usize as u32);
         // CPSR: System mode
@@ -423,22 +431,34 @@ impl State {
     }
 
     #[inline(always)]
-    pub fn is_cpu_lock_active<System: PortInstance>(&self) -> bool {
+    pub fn is_cpu_lock_active<Traits: PortInstance>(&self) -> bool {
         let cpsr: u32;
         unsafe { asm!("mrs {}, cpsr", out(reg) cpsr) };
         (cpsr & (1 << 7)) != 0
     }
 
     #[inline(always)]
-    pub fn is_task_context<System: PortInstance>(&self) -> bool {
+    pub fn is_task_context<Traits: PortInstance>(&self) -> bool {
         let cpsr: u32;
         unsafe { asm!("mrs {}, cpsr", out(reg) cpsr) };
         (cpsr & 0xf) == 0xf // System mode
     }
 
+    #[inline]
+    pub fn is_interrupt_context<Traits: PortInstance>(&self) -> bool {
+        self.is_scheduler_active::<Traits>() && !self.is_task_context::<Traits>()
+    }
+
+    #[inline]
+    pub fn is_scheduler_active<Traits: PortInstance>(&self) -> bool {
+        // `main_stack` is assigned by `dispatch_first_task`
+        // [ref:arm_main_stack_assigned_in_dft]
+        unsafe { *self.main_stack.get() != 0 }
+    }
+
     /// Implements [`crate::EntryPoint::irq_entry`]
     #[naked]
-    pub unsafe extern "C" fn irq_entry<System: PortInstance>() -> ! {
+    pub unsafe extern "C" fn irq_entry<Traits: PortInstance>() -> ! {
         unsafe {
             asm!("
                 # Adjust `lr_irq` to get the preferred return address. (The
@@ -584,32 +604,32 @@ impl State {
                 # second-level state of the next task.
                 b {push_second_level_state_and_dispatch_shortcutting}
                 ",
-                handle_irq = sym Self::handle_irq::<System>,
+                handle_irq = sym Self::handle_irq::<Traits>,
                 push_second_level_state_and_dispatch_shortcutting =
-                    sym Self::push_second_level_state_and_dispatch_shortcutting::<System>,
+                    sym Self::push_second_level_state_and_dispatch_shortcutting::<Traits>,
                 options(noreturn),
             );
         }
     }
 
-    unsafe fn handle_irq<System: PortInstance>() {
+    unsafe fn handle_irq<Traits: PortInstance>() {
         // Safety: We are the port, so it's okay to call this
-        if let Some(line) = unsafe { System::acknowledge_interrupt() } {
+        if let Some(line) = unsafe { Traits::acknowledge_interrupt() } {
             // Now that we have signaled the acknowledgement of the current
             // exception, we can start accepting nested exceptions.
             unsafe { asm!("cpsie i") };
 
-            if let Some(handler) = System::INTERRUPT_HANDLERS.get(line) {
+            if let Some(handler) = Traits::INTERRUPT_HANDLERS.get(line) {
                 // Safety: The first-level interrupt handler is the only code
                 //         allowed to call this
                 unsafe { handler() };
             }
 
             // Safety: We are the port, so it's okay to call this
-            unsafe { System::end_interrupt(line) };
+            unsafe { Traits::end_interrupt(line) };
         }
     }
 }
 
 /// Used by `use_port!`
-pub const fn validate<System: PortInstance>() {}
+pub const fn validate<Traits: PortInstance>() {}
